import { __awaiter, __decorate, __param } from "tslib";
import { FORCE_HTTP, GENESIS_SOCKET_URL } from '@genesislcap/foundation-utils';
import { observable } from '@microsoft/fast-element';
import { DI } from '@microsoft/fast-foundation';
import { scan } from 'rxjs';
import { MetaCache } from '../metadata';
import { logger } from '../utils';
import { ConnectEventsEmitter } from './connect.events';
import { ConnectConfig } from './connect.types';
import { HttpConnect } from './http.connect';
import { EventMessageType, MessageBuilder, } from './message';
import { Socket } from './socket';
import { updateState } from './updateState';
/**
 * The default implementation for the Connect interface (WS-only).
 * @public
 */
let DefaultConnect = class DefaultConnect {
    get isConnected() {
        return this.socket.isConnected;
    }
    get isConnectedSubject() {
        return this.socket.isConnectedSubject;
    }
    get isConnected$() {
        var _a;
        return (_a = this.socket.isConnectedSubject) === null || _a === void 0 ? void 0 : _a.asObservable();
    }
    get host() {
        return this.socket.host;
    }
    constructor(socket, messageBuilder, metaCache, events, config) {
        this.socket = socket;
        this.messageBuilder = messageBuilder;
        this.metaCache = metaCache;
        this.events = events;
        this.config = config;
        this.isWorking = false;
    }
    httpMode() {
        return false;
    }
    /**
     * @deprecated - Please reference {@link @genesislcap/foundation-user#User} instead.',
     */
    setValidSession(valid) {
        logger.deprecated('setValidSession', 'Please refrain from calling this API as this state is managed internally.', '15.0.0');
    }
    send(message, needsHandling) {
        return this.socket.send(message, needsHandling);
    }
    connect(host = GENESIS_SOCKET_URL) {
        return __awaiter(this, void 0, void 0, function* () {
            this.socket.reset();
            return this.socket.connect(host, this.config.connect, this.config.reconnect);
        });
    }
    disconnect() {
        this.socket.reset();
    }
    snapshot(resourceName, params) {
        return __awaiter(this, void 0, void 0, function* () {
            const snapshotMessage = this.messageBuilder.createDataLogonMessage(resourceName, params);
            logger.debug(`${snapshotMessage.MESSAGE_TYPE} (snapshot) message created -> `, snapshotMessage);
            return this.socket.send(snapshotMessage);
        });
    }
    stream(resourceName, onMessage, onError, params) {
        const msg = this.messageBuilder.createDataLogonMessage(resourceName, params);
        logger.debug(`${msg.MESSAGE_TYPE} (stream) message created -> `, msg);
        const resource = this.socket.sendForStream(msg, onMessage, (error) => {
            onError(error);
            this.events.onStreamErrorEvent(resourceName, resource, error);
        }, () => {
            this.events.onStreamCompleteEvent(resourceName, resource);
        });
        this.events.onStreamEvent(resourceName, msg, resource, {
            onMessage,
            onError,
        });
        return resource;
    }
    streamState(resourceName, onMessage = () => { }, onError = () => { }, params, initialState = []) {
        return this.stream(resourceName, onMessage, onError, params).pipe(scan(updateState, initialState));
    }
    streamWithoutAutoTeardown(resourceName, onMessage, onError, params) {
        logger.deprecated('streamWithoutAutoTeardown', `Please use 'stream' api instead.`, '15.0.0');
        const msg = this.messageBuilder.createDataLogonMessage(resourceName, params);
        logger.debug(`${msg.MESSAGE_TYPE} (stream without auto-teardown) message created -> `, msg);
        const resource = this.socket.sendForStreamWithoutTeardown(msg, onMessage, (error) => {
            onError(error);
            this.events.onStreamErrorEvent(resourceName, resource, error);
        });
        this.events.onStreamEvent(resourceName, msg, resource, {
            onMessage,
            onError,
        });
        return resource;
    }
    getMoreRows(sourceRef) {
        return __awaiter(this, void 0, void 0, function* () {
            const moreRowsMessage = this.messageBuilder.createMoreRowsMessage(sourceRef);
            logger.debug(`${moreRowsMessage.MESSAGE_TYPE} message created -> `, moreRowsMessage);
            return this.socket.send(moreRowsMessage, false);
        });
    }
    getMoreColumns(sourceRef) {
        return __awaiter(this, void 0, void 0, function* () {
            const message = this.messageBuilder.createMoreColumnsMessage(sourceRef);
            return this.socket.send(message);
        });
    }
    dataLogoff(streamSourceRef) {
        return __awaiter(this, void 0, void 0, function* () {
            const dataLogoffMessage = this.messageBuilder.createDataLogoffMessage(streamSourceRef);
            logger.debug(`${dataLogoffMessage.MESSAGE_TYPE} message created -> `, dataLogoffMessage);
            return this.socket.send(dataLogoffMessage);
        });
    }
    getMetadata(resourceName, useCache = true) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            if (useCache) {
                if (this.metaCache.hasMetadataFor(resourceName)) {
                    logger.debug(`Cached metadata retrieved for ${resourceName}`);
                    return this.metaCache.getMetadataFor(resourceName);
                }
            }
            const metaRequestMessage = this.messageBuilder.createMetaRequestMessage(resourceName);
            logger.debug(`${metaRequestMessage.MESSAGE_TYPE} (metadata) message created -> `, metaRequestMessage);
            let details;
            let error;
            try {
                this.isWorking = true;
                const metaResponse = yield this.socket.send(metaRequestMessage);
                details = (_a = metaResponse === null || metaResponse === void 0 ? void 0 : metaResponse.DETAILS) !== null && _a !== void 0 ? _a : {};
                if (Object.keys(details).length) {
                    this.metaCache.setMetadataFor(resourceName, details);
                }
            }
            catch (e) {
                error = e;
                logger.error(e);
                throw error;
            }
            finally {
                this.events.onMetadataEvent(resourceName, metaRequestMessage, error);
                this.isWorking = false;
            }
            return details;
        });
    }
    getJSONSchema(resourceName, useCache = true) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            if (useCache) {
                if (this.metaCache.hasJsonSchemaFor(resourceName)) {
                    logger.debug(`Cached JsonSchema retrieved for ${resourceName}`);
                    return this.metaCache.getJsonSchemaFor(resourceName);
                }
            }
            const metaRequestMessage = this.messageBuilder.createMetaRequestMessage(resourceName, EventMessageType.JSON_SCHEMA_REQUEST);
            logger.debug(`${metaRequestMessage.MESSAGE_TYPE} (json schema) message created -> `, metaRequestMessage);
            let jsonSchema;
            try {
                this.isWorking = true;
                jsonSchema = (_a = (yield this.socket.send(metaRequestMessage))) !== null && _a !== void 0 ? _a : {};
                if (Object.keys(jsonSchema).length) {
                    this.metaCache.setJsonSchemaFor(resourceName, jsonSchema);
                }
            }
            catch (e) {
                logger.error(e);
            }
            finally {
                this.isWorking = false;
            }
            return jsonSchema;
        });
    }
    request(resourceName, params) {
        const requestMessage = this.messageBuilder.createRequestMessage(resourceName, params);
        logger.debug(`${requestMessage.MESSAGE_TYPE} (request server) message created -> `, requestMessage);
        return this.socket.send(requestMessage);
    }
    getAvailableResources(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const requestMessage = this.messageBuilder.createResourcesMessage(params);
            logger.debug(`${requestMessage.MESSAGE_TYPE} (resources) message created -> `, requestMessage);
            return this.socket.send(requestMessage);
        });
    }
    commitEvent(eventName, params) {
        return __awaiter(this, void 0, void 0, function* () {
            const msg = this.messageBuilder.createCommitMessage(eventName, params);
            logger.debug(`${msg.MESSAGE_TYPE} (commit) message created -> `, msg);
            const promise = this.socket.send(msg);
            this.events.onCommitEvent(eventName, msg);
            return promise;
        });
    }
};
__decorate([
    observable
], DefaultConnect.prototype, "isWorking", void 0);
DefaultConnect = __decorate([
    __param(0, Socket),
    __param(1, MessageBuilder),
    __param(2, MetaCache),
    __param(3, ConnectEventsEmitter),
    __param(4, ConnectConfig)
], DefaultConnect);
export { DefaultConnect };
const useHttpConnect = FORCE_HTTP === 'true' || false;
/**
 * The DI token for the Connect interface (WS-only).
 * @internal
 */
export const WSConnect = DI.createInterface((x) => x.singleton(DefaultConnect));
/**
 * The DI token for the Connect interface (HTTP | WS).
 * @internal
 */
export const Connect = DI.createInterface((x) => useHttpConnect ? x.aliasTo(HttpConnect) : x.aliasTo(WSConnect));
/**
 * Get Connect from the DI container.
 *
 * @remarks
 * A utility method for host applications that are not using decorators or the DI container.
 *
 * @example
 * ```ts
 * import { getConnect } from '@genesislcap/foundation-comms';
 * import { API_HOST } from '@genesislcap/foundation-utils';
 * ...
 * private connect = getConnect();
 * ...
 * const connected = await this.connect.connect(API_HOST); // < WebSocket url
 * ```
 *
 * @public
 */
export function getConnect() {
    return DI.getOrCreateDOMContainer().get(Connect);
}
